/*
 * window.h: header file for window.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 *
 * @(#)$Id: window.h,v 1.6 1994/07/02 02:38:10 mrg Exp $
 */

#ifndef _WINDOW_H_
#define _WINDOW_H_
#define IN_WINDOW_H

struct WindowStru; /*XXXX*/

#include "irc_std.h"

#include "edit.h"
#include "hold.h"
#include "lastlog.h"

/* used by the update flag to determine what needs updating */
#define REDRAW_DISPLAY_FULL 1
#define REDRAW_DISPLAY_FAST 2
#define UPDATE_STATUS 4
#define REDRAW_STATUS 8

/* NickList: structure for the list of nicknames of people on a channel */
typedef struct nick_stru
{
	struct	nick_stru	*next;	/* pointer to next nickname entry */
	char	*nick;			/* nickname of person on channel */
	int	chanop;			/* True if the given nick has chanop */
}	NickList;

typedef	struct	DisplayStru
{
	char	*line;
	int	linetype;
	struct	DisplayStru	*next;
}	Display;

#define	LT_UNLOGGED	0
#define	LT_LOGHEAD	1
#define	LT_LOGTAIL	2

struct	ScreenStru;	/* ooh! */

typedef	struct	WindowStru
{
	unsigned refnum;		/* the unique reference number,
					 * assigned by IRCII */
	char	*name;			/* window's logical name */
	int	server;			/* server index */
	int	top;			/* The top line of the window, screen
					 * coordinates */
	int	bottom;			/* The botton line of the window, screen
					 * coordinates */
	int	cursor;			/* The cursor position in the window, window
					 * relative coordinates */
	int	line_cnt;		/* counter of number of lines displayed in
					 * window */
	int	scroll;			/* true, window scrolls... false window wraps */
	int	display_size;		/* number of lines of window */
	int	old_size;		/* if new_size != display_size,
					 * resize_display is called */
	int	visible;		/* true, window is drawn... false window is
					 * hidden */
	int	update;			/* window needs updating flag */
	unsigned miscflags;		/* Miscellaneous flags. */
	int	beep_always;		/* Beep even when hidden */

	char	*prompt;		/* A prompt string, usually set by EXEC'd process */
	char	*status_line[2];	/* The status line string */
	int	double_status;		/* Whether to display the 2nd status */

	Display	*top_of_display,	/* Pointer to first line of display structure */
		*display_ip;		/* Pointer to insertiong point of display
					 * structure */

	char	*current_channel;	/* Window's current channel */
        char    *waiting_channel;       /* The channel that you _want_ to be on
                                         * this window .. derived from a real
                                         * fix for the phone bug */
        char    *bind_channel;          /* the window's bound channel */
	char	*query_nick;		/* User being QUERY'ied in this window */
	NickList	*nicks;		/* List of nicks that will go to window */
	int	window_level;		/* The LEVEL of the window, determines what
					 * messages go to it */

	/* hold stuff */
	int	hold_mode;		/* true, hold mode is on for window...
					 * false it isn't */
	int	hold_on_next_rite;	/* true, the next call to rite() will
					 * activate a hold */
	int	held;			/* true, the window is currently being
					 * held */
	int	last_held;		/* Previous value of hold flag.  Used
					 * for various updating needs */
	Hold	*hold_head,		/* Pointer to first entry in hold
					 * list */
		*hold_tail;		/* Pointer to last entry in hold list */
	int	held_lines;		/* number of lines being held */
	int	scrolled_lines;		/* number of lines scrolled back */
	int	new_scrolled_lines;	/* number of lines since scroll back
					 * keys where pressed */

	/* lastlog stuff */
	Lastlog	*lastlog_head;		/* pointer to top of lastlog list */
	Lastlog	*lastlog_tail;		/* pointer to bottom of lastlog list */
	int	lastlog_level;		/* The LASTLOG_LEVEL, determines what
					 * messages go to lastlog */
	int	lastlog_size;		/* Max number of messages for the window
					 * lastlog */

	int	notify_level;		/* the notify level.. */
	char	*logfile;		/* window's logfile name */
	/* window log stuff */
	int	log;			/* true, file logging for window is on */
	FILE	*log_fp;		/* file pointer for the log file */

	struct	ScreenStru	*screen;

	struct	WindowStru	*next;	/* pointer to next entry in window list (null
					 * is end) */
	struct	WindowStru	*prev;	/* pointer to previous entry in window list
					 * (null is end) */
}	Window;

/*
 * WindowStack: The structure for the POP, PUSH, and STACK functions. A
 * simple linked list with window refnums as the data 
 */
typedef	struct	window_stack_stru
{
	unsigned int	refnum;
	struct	window_stack_stru	*next;
}	WindowStack;

extern	Window	*to_window;

extern	Window	*invisible_list;
extern	char	underline;
extern	int	who_level;
extern	char	*who_from;
extern	int	in_window_command;

typedef	struct
{
	int	top;
	int	bottom;
	int	position;
}	ShrinkInfo;

extern	void	set_scroll_lines 	(int);
extern	void	set_scroll 		(int);
extern	void	reset_line_cnt 		(int);
extern	void	set_continued_line 	(char *);
extern	void	set_underline_video 	(int);
extern	int	rite 			(Window *, char *, int, int, int, int);
extern	void	erase_display 		(Window *);
extern	ShrinkInfo	resize_display 	(Window *);
extern	void	redraw_all_windows 	(void);
extern	void	add_to_screen 		(char *);
extern	int	unhold_windows 		(void);
extern	Window	*traverse_all_windows 	(int *);
extern	void	add_to_invisible_list 	(Window *);
extern	void	delete_window 		(Window *);

/* var_settings indexes */
#define OFF 0
#define ON 1
#define TOGGLE 2

extern	Window	*add_to_window_list 	(Window *);
extern	void	erase_display 		(Window *);
extern	void	set_scroll 		(int);
extern	void	set_scroll_lines 	(int);
extern	void	update_all_status 	(void);
extern	void	set_query_nick		(char *);
extern	char	*query_nick 		(void);
extern	void	update_window_status 	(Window *, int);
extern	void	windowcmd		(char *, char *, char *);
extern	void	redraw_window 		(Window *, int);
extern	void	redraw_all_windows 	(void);
extern	void	next_window 		(char, char *);
extern	void	swap_last_window 	(char, char *);
extern	void	swap_next_window 	(char, char *);
extern	void	previous_window 	(char, char *);
extern	void	swap_previous_window 	(char, char *);
extern	void	unstop_all_windows 	(char, char *);
extern	void	back_window 		(void);
extern	void	window_kill_swap 	(void);
extern	int	is_current_channel 	(char *, char);
extern	void	redraw_all_status 	(void);
extern	void	message_to 		(unsigned);
extern	void	message_from 		(char *, int);
extern	int	get_visible_by_refnum 	(char *);
extern	unsigned int	window_display;

extern	int	is_main_screen 		(struct ScreenStru *);
extern	void	kill_screen 		(struct ScreenStru *);

extern	void	set_prompt_by_refnum 	(unsigned, char *);
extern	int	number_of_windows 	(void);
extern	void	clear_window_by_refnum 	(unsigned);
extern	unsigned current_refnum 	(void);
extern	Window	*get_window_by_refnum 	(unsigned);
extern	char	*get_target_by_refnum 	(unsigned);
extern	char	*get_prompt_by_refnum 	(unsigned);
extern	char	*set_channel_by_refnum 	(unsigned, char *);
extern	char	*get_channel_by_refnum 	(unsigned);
extern	void	set_window_server 	(int, int, int);
extern	Window	*get_window_by_name 	(char *);
extern	void	window_redirect 	(char *, int);
extern	int	get_window_server 	(unsigned);
extern	int	message_from_level 	(int);
extern	void	restore_message_from 	(void);
extern	void	save_message_from 	(void);
extern	void	window_check_servers 	(void);
extern	void	set_current_window 	(Window *);
extern	void	set_level_by_refnum 	(unsigned, int);
extern  int     is_bound 		(char *, int);
extern  void    unbind_channel 		(char *, int);
extern  char    *get_bound_channel 	(Window *);

extern	int	is_bound_to_window 	(Window *, char *);
extern	int	is_bound_anywhere 	(char *);

#define WINDOW_NOTIFY	((unsigned) 0x0001)
#define WINDOW_NOTIFIED	((unsigned) 0x0002)

#undef IN_WINDOW_H
#endif /* _WINDOW_H_ */
