/*
 * mail.c -- a gross simplification of mail checking.
 * Only unix maildrops are supported.
 *
 * Written by Jeremy Nelson
 * Copyright 1996 EPIC Software Labs
 */

#include "irc.h"
#include "mail.h"
#include "lastlog.h"
#include "hook.h"
#include "vars.h"
#include "ircaux.h"
#include "output.h"

static	char	*mail_path = (char *) 0;
static	int	mail_latch = 0;

int check_mail_status (void)
{
	int fs;

	if (!get_int_var(MAIL_VAR))
		return 0;

	if (!mail_path)
	{
		char *tmp_mail_path;

		if ((tmp_mail_path = getenv("MAIL")) != NULL)
			mail_path = m_strdup(tmp_mail_path);
		else
			mail_path = m_3dup(UNIX_MAIL, "/", username);
	}

	fs = file_size(mail_path);
	return (fs <= 0) ? 0 : fs;
}


/*
 * check_mail: returns NULL if there is no outstanding mail, return the 
 * empty_string if there is outstanding mail.  Each time it is detected 
 * that there is email when there has previously been none, the /on MAIL
 * hook will be thrown.  It is no longer possible to determine how many
 * outstanding emails are available, as that feature was apparantly not
 * appreciated by many people.
 */
char	*check_mail (void)
{
	if (check_mail_status())
	{
		if (!mail_latch)
		{
			mail_latch = 1;
			if (do_hook(MAIL_LIST, "You have new email"))
			{
				int lastlog_level = set_lastlog_msg_level(LOG_CRAP);
				say("You have new email.");
				set_lastlog_msg_level(lastlog_level);
			}
		}
		return empty_string;
	}

	if (mail_latch)
		mail_latch = 0;

	return NULL;
}
